class MessageKeys:
    def __init__(
        self, cipher_key: bytes, mac_key: bytes, iv: bytes, counter: int
    ) -> None:
        self._cipher_key = cipher_key
        self._mac_key = mac_key
        self._iv = iv
        self._counter = counter

    def get_cipher_key(self) -> bytes:
        return self._cipher_key

    def get_mac_key(self) -> bytes:
        return self._mac_key

    def get_iv(self) -> bytes:
        return self._iv

    def get_counter(self) -> int:
        return self._counter



from __future__ import annotations

from typing import cast

import google.protobuf.message
from google.protobuf.message import DecodeError

from ..ecc.curve import Curve
from ..ecc.djbec import CurvePublicKey
from ..exceptions import InvalidKeyException
from ..exceptions import InvalidMessageException
from ..exceptions import InvalidVersionException
from ..identitykey import IdentityKey
from ..util.byteutil import ByteUtil
from . import whisper_pb2
from .ciphertextmessage import CiphertextMessage
from .whispermessage import WhisperMessage


class PreKeyWhisperMessage(CiphertextMessage):
    def __init__(
        self,
        serialized: bytes,
        message_version: int,
        device_id: int,
        pre_key_id: int,
        signed_pre_key_id: int,
        ec_public_base_key: CurvePublicKey,
        identity_key: IdentityKey,
        whisper_message: WhisperMessage,
    ) -> None:
        self._serialized = serialized
        self._message_version = message_version
        self._device_id = device_id
        self._pre_key_id = pre_key_id
        self._signed_pre_key_id = signed_pre_key_id
        self._ec_public_base_key = ec_public_base_key
        self._identity_key = identity_key
        self._whisper_message = whisper_message

    @classmethod
    def new(
        cls,
        message_version: int,
        device_id: int,
        pre_key_id: int,
        signed_pre_key_id: int,
        ec_public_base_key: CurvePublicKey,
        identity_key: IdentityKey,
        whisper_message: WhisperMessage,
    ) -> PreKeyWhisperMessage:
        prekey_message = cast(
            PreKeyWhisperMessageProto,
            whisper_pb2.PreKeyWhisperMessage(),  # pyright: ignore
        )
        prekey_message.signedPreKeyId = signed_pre_key_id
        prekey_message.preKeyId = pre_key_id
        prekey_message.baseKey = ec_public_base_key.serialize()
        prekey_message.identityKey = identity_key.serialize()
        prekey_message.message = whisper_message.serialize()
        prekey_message.registrationId = device_id

        version_bytes = ByteUtil.ints_to_byte_high_and_low(3, 3)
        message_bytes = prekey_message.SerializeToString()

        serialized = ByteUtil.combine(version_bytes, message_bytes)

        return cls(
            serialized,
            message_version,
            device_id,
            pre_key_id,
            signed_pre_key_id,
            ec_public_base_key,
            identity_key,
            whisper_message,
        )

    @classmethod
    def from_bytes(cls, serialized: bytes) -> PreKeyWhisperMessage:
        try:
            version = ByteUtil.high_bits_to_int(serialized[0])
            if version != 3:
                raise InvalidVersionException("Unknown version %s" % version)

            pre_key_whisper_message = cast(
                PreKeyWhisperMessageProto,
                whisper_pb2.PreKeyWhisperMessage(),  # pyright: ignore
            )
            pre_key_whisper_message.ParseFromString(serialized[1:])

            if (
                not pre_key_whisper_message.signedPreKeyId
                or not pre_key_whisper_message.baseKey
                or not pre_key_whisper_message.identityKey
                or not pre_key_whisper_message.message
            ):
                raise InvalidMessageException("Incomplete message")

            device_id = pre_key_whisper_message.registrationId
            pre_key_id = pre_key_whisper_message.preKeyId
            signed_pre_key_id = pre_key_whisper_message.signedPreKeyId

            base_key = Curve.decode_point(pre_key_whisper_message.baseKey)

            identity_key = IdentityKey(
                Curve.decode_point(pre_key_whisper_message.identityKey)
            )
            message = WhisperMessage.from_bytes(pre_key_whisper_message.message)
        except (InvalidKeyException, DecodeError) as error:
            raise InvalidMessageException(str(error))

        return cls(
            serialized,
            version,
            device_id,
            pre_key_id,
            signed_pre_key_id,
            base_key,
            identity_key,
            message,
        )

    def get_message_version(self) -> int:
        return self._message_version

    def get_identity_key(self) -> IdentityKey:
        return self._identity_key

    def get_device_id(self) -> int:
        return self._device_id

    def get_pre_key_id(self) -> int:
        return self._pre_key_id

    def get_signed_pre_key_id(self) -> int:
        return self._signed_pre_key_id

    def get_base_key(self) -> CurvePublicKey:
        return self._ec_public_base_key

    def get_whisper_message(self) -> WhisperMessage:
        return self._whisper_message

    def serialize(self) -> bytes:
        return self._serialized

    def get_type(self) -> int:
        return CiphertextMessage.PREKEY_TYPE


class PreKeyWhisperMessageProto(google.protobuf.message.Message):
    signedPreKeyId: int
    preKeyId: int
    baseKey: bytes
    identityKey: bytes
    message: bytes
    registrationId: int



from __future__ import annotations

from ..ecc.djbec import CurvePublicKey
from ..ecc.eckeypair import ECKeyPair
from ..identitykey import IdentityKey
from ..identitykeypair import IdentityKeyPair


class AliceParameters:
    def __init__(
        self,
        our_identity_key: IdentityKeyPair,
        our_base_key: ECKeyPair,
        their_identity_key: IdentityKey,
        their_signed_pre_key: CurvePublicKey,
        their_ratchet_key: CurvePublicKey,
        their_one_time_pre_key: CurvePublicKey,
    ) -> None:
        self._our_base_key = our_base_key
        self._our_identity_key = our_identity_key
        self._their_signed_pre_key = their_signed_pre_key
        self._their_ratchet_key = their_ratchet_key
        self._their_identity_key = their_identity_key
        self._their_one_time_pre_key = their_one_time_pre_key

    def get_our_identity_key(self) -> IdentityKeyPair:
        return self._our_identity_key

    def get_our_base_key(self) -> ECKeyPair:
        return self._our_base_key

    def get_their_identity_key(self) -> IdentityKey:
        return self._their_identity_key

    def get_their_signed_pre_key(self) -> CurvePublicKey:
        return self._their_signed_pre_key

    def get_their_one_time_pre_key(self) -> CurvePublicKey:
        return self._their_one_time_pre_key

    def get_their_ratchet_key(self) -> CurvePublicKey:
        return self._their_ratchet_key



from __future__ import annotations

from typing import cast
from typing import Optional
from typing import Union

import google.protobuf.message

from ..ecc.curve import Curve
from ..ecc.djbec import CurvePublicKey
from ..ecc.eckeypair import ECKeyPair
from ..identitykeypair import IdentityKey
from ..kdf.messagekeys import MessageKeys
from ..ratchet.chainkey import ChainKey
from ..ratchet.rootkey import RootKey
from . import storage_pb2


class SessionState:
    def __init__(
        self, session: Optional[Union[SessionState, SessionStructureProto]] = None
    ) -> None:
        if session is None:
            self._session_structure = cast(
                SessionStructureProto, storage_pb2.SessionStructure()  # pyright: ignore
            )

        elif isinstance(session, SessionState):
            self._session_structure = cast(
                SessionStructureProto, storage_pb2.SessionStructure()  # pyright: ignore
            )
            self._session_structure.CopyFrom(session.get_structure())

        else:
            self._session_structure = session

    def get_structure(self) -> SessionStructureProto:
        return self._session_structure

    def get_alice_base_key(self) -> bytes:
        return self._session_structure.aliceBaseKey

    def set_alice_base_key(self, alice_base_key: bytes) -> None:
        self._session_structure.aliceBaseKey = alice_base_key

    def set_session_version(self, version: int) -> None:
        self._session_structure.sessionVersion = version

    def get_session_version(self) -> int:
        return self._session_structure.sessionVersion

    def set_remote_identity_key(self, identity_key: IdentityKey) -> None:
        self._session_structure.remoteIdentityPublic = identity_key.serialize()

    def set_local_identity_key(self, identity_key: IdentityKey) -> None:
        self._session_structure.localIdentityPublic = identity_key.serialize()

    def get_remote_identity_key(self) -> IdentityKey:
        assert self._session_structure.remoteIdentityPublic is not None
        return IdentityKey(self._session_structure.remoteIdentityPublic)

    def get_local_identity_key(self) -> IdentityKey:
        return IdentityKey(self._session_structure.localIdentityPublic)

    def get_previous_counter(self) -> int:
        return self._session_structure.previousCounter

    def set_previous_counter(self, previous_counter: int) -> None:
        self._session_structure.previousCounter = previous_counter

    def get_root_key(self) -> RootKey:
        return RootKey(
            self.get_session_version(),
            self._session_structure.rootKey,
        )

    def set_root_key(self, root_key: RootKey) -> None:
        self._session_structure.rootKey = root_key.get_key_bytes()

    def get_sender_ratchet_key(self) -> CurvePublicKey:
        key = Curve.decode_point(self._session_structure.senderChain.senderRatchetKey)
        assert isinstance(key, CurvePublicKey)
        return key

    def get_sender_ratchet_key_pair(self) -> ECKeyPair:
        public_key = self.get_sender_ratchet_key()
        private_key = Curve.decode_private_point(
            self._session_structure.senderChain.senderRatchetKeyPrivate
        )

        return ECKeyPair(public_key, private_key)

    def has_receiver_chain(
        self, ec_publick_key_sender_ephemeral: CurvePublicKey
    ) -> bool:
        return self.get_receiver_chain(ec_publick_key_sender_ephemeral) is not None

    def has_sender_chain(self) -> bool:
        return self._session_structure.HasField("senderChain")

    def get_receiver_chain(
        self, ec_publick_key_sender_ephemeral: CurvePublicKey
    ) -> Optional[tuple[ChainStructureProto, int]]:
        receiver_chains = self._session_structure.receiverChains

        for index, receiver_chain in enumerate(receiver_chains):
            chain_sender_ratchet_key = Curve.decode_point(
                receiver_chain.senderRatchetKey
            )
            if chain_sender_ratchet_key == ec_publick_key_sender_ephemeral:
                return (receiver_chain, index)

    def get_receiver_chain_key(
        self, ec_public_key_sender_ephemeral: CurvePublicKey
    ) -> ChainKey:
        receiver_chain_and_index = self.get_receiver_chain(
            ec_public_key_sender_ephemeral
        )
        assert receiver_chain_and_index is not None
        receiver_chain = receiver_chain_and_index[0]
        assert receiver_chain is not None

        return ChainKey(
            self.get_session_version(),
            receiver_chain.chainKey.key,
            receiver_chain.chainKey.index,
        )

    def add_receiver_chain(
        self, ec_publick_key_sender_ratchet_key: CurvePublicKey, chain_key: ChainKey
    ) -> None:
        sender_ratchet_key = ec_publick_key_sender_ratchet_key

        chain = cast(
            ChainStructureProto,
            storage_pb2.SessionStructure.Chain(),  # pyright: ignore
        )
        chain.senderRatchetKey = sender_ratchet_key.serialize()
        chain.chainKey.key = chain_key.get_key()
        chain.chainKey.index = chain_key.get_index()

        self._session_structure.receiverChains.extend([chain])

        if len(self._session_structure.receiverChains) > 5:
            del self._session_structure.receiverChains[0]

    def set_sender_chain(
        self, ec_key_pair_sender_ratchet_key_pair: ECKeyPair, chain_key: ChainKey
    ) -> None:
        sender_ratchet_key_pair = ec_key_pair_sender_ratchet_key_pair

        self._session_structure.senderChain.senderRatchetKey = (
            sender_ratchet_key_pair.get_public_key().serialize()
        )
        self._session_structure.senderChain.senderRatchetKeyPrivate = (
            sender_ratchet_key_pair.get_private_key().serialize()
        )
        self._session_structure.senderChain.chainKey.key = chain_key.get_key()
        self._session_structure.senderChain.chainKey.index = chain_key.get_index()

    def get_sender_chain_key(self) -> ChainKey:
        chain_key_structure = self._session_structure.senderChain.chainKey
        return ChainKey(
            self.get_session_version(),
            chain_key_structure.key,
            chain_key_structure.index,
        )

    def set_sender_chain_key(self, chain_key_next_chain_key: ChainKey) -> None:
        next_chain_key = chain_key_next_chain_key

        self._session_structure.senderChain.chainKey.key = next_chain_key.get_key()
        self._session_structure.senderChain.chainKey.index = next_chain_key.get_index()

    def has_message_keys(
        self, ec_publick_key_sender_ephemeral: CurvePublicKey, counter: int
    ) -> bool:
        sender_ephemeral = ec_publick_key_sender_ephemeral
        chain_and_index = self.get_receiver_chain(sender_ephemeral)
        assert chain_and_index is not None
        chain = chain_and_index[0]

        message_key_list = chain.messageKeys
        for message_key in message_key_list:
            if message_key.index == counter:
                return True

        return False

    def remove_message_keys(
        self, ec_public_key_sender_ephemeral: CurvePublicKey, counter: int
    ) -> MessageKeys:
        sender_ephemeral = ec_public_key_sender_ephemeral
        chain_and_index = self.get_receiver_chain(sender_ephemeral)
        assert chain_and_index is not None
        chain = chain_and_index[0]
        assert chain is not None

        message_key_list = chain.messageKeys
        result = None

        for i in range(0, len(message_key_list)):
            message_key = message_key_list[i]
            if message_key.index == counter:
                result = MessageKeys(
                    message_key.cipherKey,
                    message_key.macKey,
                    message_key.iv,
                    message_key.index,
                )
                del message_key_list[i]
                break

        assert result is not None

        self._session_structure.receiverChains[chain_and_index[1]].CopyFrom(chain)

        return result

    def set_message_keys(
        self, ec_public_key_sender_ephemeral: CurvePublicKey, message_keys: MessageKeys
    ) -> None:
        sender_ephemeral = ec_public_key_sender_ephemeral
        chain_and_index = self.get_receiver_chain(sender_ephemeral)
        assert chain_and_index is not None
        chain = chain_and_index[0]
        message_key_structure = chain.messageKeys.add()  # pyright: ignore
        message_key_structure.cipherKey = message_keys.get_cipher_key()
        message_key_structure.macKey = message_keys.get_mac_key()
        message_key_structure.index = message_keys.get_counter()
        message_key_structure.iv = message_keys.get_iv()

        self._session_structure.receiverChains[chain_and_index[1]].CopyFrom(chain)

    def set_receiver_chain_key(
        self, ec_public_key_sender_ephemeral: CurvePublicKey, chain_key: ChainKey
    ) -> None:
        sender_ephemeral = ec_public_key_sender_ephemeral
        chain_and_index = self.get_receiver_chain(sender_ephemeral)
        assert chain_and_index is not None
        chain = chain_and_index[0]
        chain.chainKey.key = chain_key.get_key()
        chain.chainKey.index = chain_key.get_index()

        self._session_structure.receiverChains[chain_and_index[1]].CopyFrom(chain)

    def set_unacknowledged_pre_key_message(
        self, pre_key_id: int, signed_pre_key_id: int, base_key: CurvePublicKey
    ) -> None:
        self._session_structure.pendingPreKey.signedPreKeyId = signed_pre_key_id
        self._session_structure.pendingPreKey.baseKey = base_key.serialize()
        self._session_structure.pendingPreKey.preKeyId = pre_key_id

    def has_unacknowledged_pre_key_message(self) -> bool:
        return self._session_structure.HasField("pendingPreKey")

    def get_unacknowledged_pre_key_message_items(
        self,
    ) -> UnacknowledgedPreKeyMessageItems:
        pre_key_id = None
        if self._session_structure.pendingPreKey.HasField("preKeyId"):
            pre_key_id = self._session_structure.pendingPreKey.preKeyId

        assert pre_key_id is not None
        return UnacknowledgedPreKeyMessageItems(
            pre_key_id,
            self._session_structure.pendingPreKey.signedPreKeyId,
            Curve.decode_point(self._session_structure.pendingPreKey.baseKey),
        )

    def clear_unacknowledged_pre_key_message(self) -> None:
        self._session_structure.ClearField("pendingPreKey")

    def set_remote_device_id(self, device_id: int) -> None:
        self._session_structure.remoteRegistrationId = device_id

    def get_remote_device_id(self) -> int:
        return self._session_structure.remoteRegistrationId

    def set_our_device_id(self, device_id: int) -> None:
        self._session_structure.localRegistrationId = device_id

    def get_our_device_id(self) -> int:
        return self._session_structure.localRegistrationId

    def serialize(self) -> bytes:
        return self._session_structure.SerializeToString()


class UnacknowledgedPreKeyMessageItems:
    def __init__(
        self, pre_key_id: int, signed_pre_key_id: int, base_key: CurvePublicKey
    ) -> None:
        self._pre_key_id = pre_key_id
        self._signed_pre_key_id = signed_pre_key_id
        self._base_key = base_key

    def get_pre_key_id(self) -> int:
        return self._pre_key_id

    def get_signed_pre_key_id(self) -> int:
        return self._signed_pre_key_id

    def get_base_key(self) -> CurvePublicKey:
        return self._base_key


class MessageKeyStructureProto(google.protobuf.message.Message):
    index: int
    cipherKey: bytes
    macKey: bytes
    iv: bytes


class ChainKeyStructureProto(google.protobuf.message.Message):
    index: int
    key: bytes


class ChainStructureProto(google.protobuf.message.Message):
    senderRatchetKey: bytes
    senderRatchetKeyPrivate: bytes
    chainKey: ChainKeyStructureProto
    messageKeys: list[MessageKeyStructureProto]


class PendingKeyExchangeStructureProto(google.protobuf.message.Message):
    sequence: int
    localBaseKey: bytes
    localBaseKeyPrivate: bytes
    localRatchetKey: bytes
    localRatchetKeyPrivate: bytes
    localIdentityKey: bytes
    localIdentityKeyPrivate: bytes


class PendingPreKeyStructureProto(google.protobuf.message.Message):
    preKeyId: int
    signedPreKeyId: int
    baseKey: bytes


class SessionStructureProto(google.protobuf.message.Message):
    sessionVersion: int
    localIdentityPublic: bytes
    remoteIdentityPublic: bytes
    rootKey: bytes
    previousCounter: int
    senderChain: ChainStructureProto
    receiverChains: list[ChainStructureProto]
    pendingKeyExchange: PendingKeyExchangeStructureProto
    pendingPreKey: PendingPreKeyStructureProto
    remoteRegistrationId: int
    localRegistrationId: int
    needsRefresh: bool
    aliceBaseKey: bytes



from .. import _curve  # pyright: ignore


def calculate_signature(random: bytes, privatekey: bytes, message: bytes) -> bytes:
    return _curve.calculateSignature(random, privatekey, message)  # pyright: ignore


def verify_signature_curve(publickey: bytes, message: bytes, signature: bytes) -> int:
    return _curve.verifySignatureCurve(publickey, message, signature)  # pyright: ignore


def verify_signature_ed(publickey: bytes, message: bytes, signature: bytes) -> int:
    return _curve.verifySignatureEd(publickey, message, signature)  # pyright: ignore


def generate_private_key(random: bytes) -> bytes:
    return _curve.generatePrivateKey(random)  # pyright: ignore


def generate_public_key(privatekey: bytes) -> bytes:
    return _curve.generatePublicKey(privatekey)  # pyright: ignore


def calculate_agreement(privatekey: bytes, publickey: bytes) -> bytes:
    return _curve.calculateAgreement(privatekey, publickey)  # pyright: ignore


def convert_curve_to_ed_pubkey(publickey: bytes) -> bytes:
    return _curve.convertCurveToEdPubkey(publickey)  # pyright: ignore


def convert_ed_to_curve_pubkey(publickey: bytes) -> bytes:
    return _curve.convertEdToCurvePubkey(publickey)  # pyright: ignore



from __future__ import annotations

from typing import Protocol
from typing import TypedDict

import random
from dataclasses import dataclass

from .const import OMEMOTrust
from .identitykey import IdentityKey


@dataclass
class IdentityInfo:
    active: bool
    address: str
    device_id: int
    label: str
    last_seen: float | None
    public_key: IdentityKey
    trust: OMEMOTrust


@dataclass
class OMEMOConfig:
    default_prekey_amount: int
    min_prekey_amount: int
    spk_archive_seconds: int
    spk_cycle_seconds: int
    unacknowledged_count: int


class PreKey(TypedDict):
    key: bytes
    id: int


class OMEMOBundleProto(Protocol):
    device_id: int
    ik: bytes
    namespace: str
    otpks: list[PreKey]
    spk: PreKey
    spk_signature: bytes

    def pick_prekey(self) -> PreKey:
        ...


@dataclass
class OMEMOBundle(OMEMOBundleProto):
    device_id: int
    ik: bytes
    namespace: str
    otpks: list[PreKey]
    spk: PreKey
    spk_signature: bytes

    def pick_prekey(self) -> PreKey:
        return random.SystemRandom().choice(self.otpks)


@dataclass
class OMEMOMessageProto(Protocol):
    sid: int
    iv: bytes
    keys: dict[int, tuple[bytes, bool]]
    payload: bytes | None


@dataclass
class OMEMOMessage(OMEMOMessageProto):
    sid: int
    iv: bytes
    keys: dict[int, tuple[bytes, bool]]
    payload: bytes | None



from __future__ import annotations

from .djbec import CurvePublicKey
from .djbec import DjbECPrivateKey


class ECKeyPair:
    def __init__(
        self, public_key: CurvePublicKey, private_key: DjbECPrivateKey
    ) -> None:
        self._public_key = public_key
        self._private_key = private_key

    def get_private_key(self) -> DjbECPrivateKey:
        return self._private_key

    def get_public_key(self) -> CurvePublicKey:
        return self._public_key



from __future__ import annotations

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf import hkdf


def derive(
    *,
    input_key_material: bytes,
    length: int,
    salt: bytes,
    info: bytes,
) -> bytes:
    return hkdf.HKDF(
        algorithm=hashes.SHA256(),
        length=length,
        salt=salt,
        info=info,
        backend=default_backend(),
    ).derive(input_key_material)



from __future__ import annotations

from typing import Any
from typing import Optional


class DuplicateMessageException(Exception):
    pass


class InvalidKeyException(Exception):
    pass


class InvalidKeyIdException(Exception):
    pass


class InvalidMessageException(Exception):
    def __init__(self, message: str, exceptions: Optional[list[Exception]] = None):
        if exceptions:
            message += str(exceptions[0])
        super(InvalidMessageException, self).__init__(message)


class InvalidVersionException(Exception):
    pass


class NoSessionException(Exception):
    pass


class UntrustedIdentityException(Exception):
    def __init__(self, name: str, identityKey: Any):
        self.name = name
        self.identityKey = identityKey

    def getName(self) -> str:
        return self.name

    def getIdentityKey(self) -> Any:
        return self.identityKey


class NoDevicesFound(Exception):
    pass


class NoValidSessions(Exception):
    pass


class SelfMessage(Exception):
    pass


class MessageNotForDevice(Exception):
    pass


class DecryptionFailed(Exception):
    pass


class KeyExchangeMessage(Exception):
    pass


class InvalidMessage(Exception):
    pass


class DuplicateMessage(Exception):
    pass


class BundleValidationError(Exception):
    pass



from __future__ import annotations

from typing import cast

import google.protobuf.message

from ..ecc.curve import Curve
from ..ecc.eckeypair import ECKeyPair
from .storage_pb2 import SignedPreKeyRecordStructure  # pyright: ignore


class SignedPreKeyRecord:
    def __init__(self, structure: SignedPreKeyRecordStructureProto) -> None:
        self._structure = structure

    @classmethod
    def new(
        cls, _id: int, timestamp: int, ec_key_pair: ECKeyPair, signature: bytes
    ) -> SignedPreKeyRecord:
        record = cast(SignedPreKeyRecordStructureProto, SignedPreKeyRecordStructure())

        record.id = _id
        record.publicKey = ec_key_pair.get_public_key().serialize()
        record.privateKey = ec_key_pair.get_private_key().serialize()
        record.signature = signature
        record.timestamp = timestamp

        return cls(record)

    @classmethod
    def from_bytes(cls, serialized: bytes) -> SignedPreKeyRecord:
        record = cast(SignedPreKeyRecordStructureProto, SignedPreKeyRecordStructure())
        record.ParseFromString(serialized)
        return cls(record)

    def get_id(self) -> int:
        return self._structure.id

    def get_timestamp(self) -> int:
        return self._structure.timestamp

    def get_key_pair(self) -> ECKeyPair:
        public_key = Curve.decode_point(self._structure.publicKey)
        private_key = Curve.decode_private_point(self._structure.privateKey)

        return ECKeyPair(public_key, private_key)

    def get_signature(self) -> bytes:
        return self._structure.signature

    def serialize(self) -> bytes:
        return self._structure.SerializeToString()


class SignedPreKeyRecordStructureProto(google.protobuf.message.Message):
    id: int
    publicKey: bytes
    privateKey: bytes
    signature: bytes
    timestamp: int



from __future__ import annotations

from typing import Optional


class ByteUtil:
    @staticmethod
    def combine(*args: int | list[int] | bytes | bytearray) -> bytes:
        baos = bytearray()
        for v in args:
            if isinstance(v, (list, bytes, bytearray)):
                baos.extend(v)
            else:
                baos.append(v)

        return bytes(baos)

    @staticmethod
    def split(
        inp: bytes,
        first_length: int,
        second_length: int,
        third_length: Optional[int] = None,
    ) -> list[bytes]:
        parts: list[bytes] = []
        parts.append(inp[:first_length])
        parts.append(inp[first_length : first_length + second_length])
        if third_length is not None:
            start = first_length + second_length
            end = first_length + second_length + third_length
            parts.append(inp[start:end])

        return parts

    @staticmethod
    def trim(inp: bytes, length: int) -> bytes:
        return inp[:length]

    @staticmethod
    def ints_to_byte_high_and_low(high_value: int, low_value: int) -> int:
        return ((high_value << 4 | low_value) & 0xFF) % 256

    @staticmethod
    def high_bits_to_int(value: int) -> int:
        return (value & 0xFF) >> 4



from __future__ import annotations

from typing import cast

import google.protobuf.message
from google.protobuf.message import DecodeError

from ..ecc.curve import Curve
from ..ecc.ec import ECPublicKey
from ..exceptions import InvalidKeyException
from ..exceptions import InvalidMessageException
from ..identitykey import IdentityKey
from . import omemo_pb2
from .ciphertextmessage import CiphertextMessage
from .omemo_message import OMEMOMessage


class OMEMOKeyExchange(CiphertextMessage):
    def __init__(
        self,
        serialized: bytes,
        message_version: int,
        device_id: int,
        pre_key_id: int,
        signed_pre_key_id: int,
        ec_public_base_key: ECPublicKey,
        identity_key: IdentityKey,
        omemo_message: OMEMOMessage,
    ):
        self._message_version = message_version
        self._device_id = device_id
        self._pre_key_id = pre_key_id
        self._signed_pre_key_id = signed_pre_key_id
        self._base_key = ec_public_base_key
        self._identity_key = identity_key
        self._message = omemo_message
        self._serialized = serialized

    @classmethod
    def new(
        cls,
        message_version: int,
        device_id: int,
        pre_key_id: int,
        signed_pre_key_id: int,
        ec_public_base_key: ECPublicKey,
        identity_key: IdentityKey,
        omemo_message: OMEMOMessage,
    ) -> OMEMOKeyExchange:
        key_exchange = cast(
            OMEMOKeyExchangeProto, omemo_pb2.OMEMOKeyExchange()  # pyright: ignore
        )
        key_exchange.pk_id = pre_key_id
        key_exchange.spk_id = signed_pre_key_id
        key_exchange.ik = identity_key.serialize()
        key_exchange.ek = ec_public_base_key.serialize()
        key_exchange.message = omemo_message.serialize()

        serialized = key_exchange.SerializeToString()

        return cls(
            serialized,
            4,
            0,
            pre_key_id,
            signed_pre_key_id,
            ec_public_base_key,
            identity_key,
            omemo_message,
        )

    @classmethod
    def from_bytes(cls, serialized: bytes) -> OMEMOKeyExchange:
        try:
            key_exchange = cast(
                OMEMOKeyExchangeProto, omemo_pb2.OMEMOKeyExchange()  # pyright: ignore
            )
            key_exchange.ParseFromString(serialized)

            if (
                not key_exchange.spk_id
                or not key_exchange.ek
                or not key_exchange.ik
                or not key_exchange.message
            ):
                raise InvalidMessageException("Incomplete message")

            pre_key_id = key_exchange.pk_id
            signed_pre_key_id = key_exchange.spk_id
            ec_public_base_key = Curve.decode_point(key_exchange.ek)
            identity_key = IdentityKey(Curve.decode_point(key_exchange.ik))
            omemo_message = OMEMOMessage.from_bytes(key_exchange.message)

        except (InvalidKeyException, DecodeError) as error:
            raise InvalidMessageException(str(error))

        return cls(
            serialized,
            4,
            0,
            pre_key_id,
            signed_pre_key_id,
            ec_public_base_key,
            identity_key,
            omemo_message,
        )

    def get_message_version(self) -> int:
        return 4

    def get_identity_key(self) -> IdentityKey:
        return self._identity_key

    def get_device_id(self) -> int:
        return self._device_id

    def get_pre_key_id(self) -> int:
        return self._pre_key_id

    def get_signed_pre_key_id(self) -> int:
        return self._signed_pre_key_id

    def get_base_key(self) -> ECPublicKey:
        return self._base_key

    def get_whisper_message(self) -> OMEMOMessage:
        return self._message

    def serialize(self) -> bytes:
        return self._serialized

    def get_type(self) -> int:
        return CiphertextMessage.PREKEY_TYPE


class OMEMOKeyExchangeProto(google.protobuf.message.Message):
    pk_id: int
    spk_id: int
    ik: bytes
    ek: bytes
    message: bytes



from ..util.byteutil import ByteUtil


class DerivedRootSecrets:
    SIZE = 64

    def __init__(self, okm: bytes) -> None:
        keys = ByteUtil.split(okm, 32, 32)
        self._root_key = keys[0]
        self._chain_key = keys[1]

    def get_root_key(self) -> bytes:
        return self._root_key

    def get_chain_key(self) -> bytes:
        return self._chain_key



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: omemo.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0bomemo.proto\x12\x05omemo\"I\n\x0cOMEMOMessage\x12\t\n\x01n\x18\x01 \x02(\r\x12\n\n\x02pn\x18\x02 \x02(\r\x12\x0e\n\x06\x64h_pub\x18\x03 \x02(\x0c\x12\x12\n\nciphertext\x18\x04 \x02(\x0c\"9\n\x19OMEMOAuthenticatedMessage\x12\x0b\n\x03mac\x18\x01 \x02(\x0c\x12\x0f\n\x07message\x18\x02 \x02(\x0c\"Z\n\x10OMEMOKeyExchange\x12\r\n\x05pk_id\x18\x01 \x02(\r\x12\x0e\n\x06spk_id\x18\x02 \x02(\r\x12\n\n\x02ik\x18\x03 \x02(\x0c\x12\n\n\x02\x65k\x18\x04 \x02(\x0c\x12\x0f\n\x07message\x18\x05 \x02(\x0c')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'omemo_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _OMEMOMESSAGE._serialized_start=22
  _OMEMOMESSAGE._serialized_end=95
  _OMEMOAUTHENTICATEDMESSAGE._serialized_start=97
  _OMEMOAUTHENTICATEDMESSAGE._serialized_end=154
  _OMEMOKEYEXCHANGE._serialized_start=156
  _OMEMOKEYEXCHANGE._serialized_end=246
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from typing import Optional

import abc

from ..const import OMEMOTrust
from ..identitykey import IdentityKey
from ..identitykeypair import IdentityKeyPair
from ..state.prekeyrecord import PreKeyRecord
from ..structs import IdentityInfo
from .sessionrecord import SessionRecord
from .signedprekeyrecord import SignedPreKeyRecord


class Store(abc.ABC):
    # Secret

    @abc.abstractmethod
    def set_our_identity(
        self, device_id: int, identity_key_pair: IdentityKeyPair
    ) -> None:
        pass

    @abc.abstractmethod
    def get_identity_key_pair(self) -> IdentityKeyPair:
        pass

    @abc.abstractmethod
    def get_our_device_id(self) -> int:
        pass

    # Identity

    @abc.abstractmethod
    def save_identity(self, recipient_id: str, identity_key: IdentityKey) -> None:
        pass

    @abc.abstractmethod
    def delete_identity(self, recipient_id: str, identity_key: IdentityKey) -> None:
        pass

    @abc.abstractmethod
    def is_trusted_identity(self, recipient_id: str, identity_key: IdentityKey) -> bool:
        pass

    @abc.abstractmethod
    def set_trust(
        self, recipient_id: str, identity_key: IdentityKey, trust: OMEMOTrust
    ) -> None:
        pass

    @abc.abstractmethod
    def set_identity_last_seen(
        self, recipient_id: str, identity_key: IdentityKey
    ) -> None:
        pass

    @abc.abstractmethod
    def get_identity_infos(self, recipient_ids: str | list[str]) -> list[IdentityInfo]:
        pass

    # Pre Keys

    @abc.abstractmethod
    def get_pre_key_count(self) -> int:
        pass

    @abc.abstractmethod
    def load_pre_key(self, pre_key_id: int) -> PreKeyRecord:
        pass

    @abc.abstractmethod
    def load_pending_pre_keys(self) -> list[PreKeyRecord]:
        pass

    @abc.abstractmethod
    def get_current_pre_key_id(self) -> Optional[int]:
        pass

    @abc.abstractmethod
    def store_pre_key(self, pre_key_id: int, pre_key_record: PreKeyRecord):
        pass

    @abc.abstractmethod
    def contains_pre_key(self, pre_key_id: int) -> bool:
        pass

    @abc.abstractmethod
    def remove_pre_key(self, pre_key_id: int) -> None:
        pass

    # Signed Pre Key

    @abc.abstractmethod
    def load_signed_pre_key(self, signed_pre_key_id: int) -> SignedPreKeyRecord:
        pass

    @abc.abstractmethod
    def load_signed_pre_keys(self) -> list[SignedPreKeyRecord]:
        pass

    @abc.abstractmethod
    def store_signed_pre_key(
        self, signed_pre_key_id: int, signed_pre_key_record: SignedPreKeyRecord
    ) -> None:
        pass

    @abc.abstractmethod
    def contains_signed_pre_key(self, signed_pre_key_id: int) -> bool:
        pass

    @abc.abstractmethod
    def get_current_signed_pre_key_id(self) -> int:
        pass

    @abc.abstractmethod
    def get_signed_pre_key_timestamp(self, signed_pre_key_id: int) -> int:
        pass

    @abc.abstractmethod
    def remove_old_signed_pre_keys(self, timestamp: int) -> None:
        pass

    @abc.abstractmethod
    def remove_signed_pre_key(self, signed_pre_key_id: int) -> None:
        pass

    # Session

    @abc.abstractmethod
    def load_session(self, recipient_id: str, device_id: int) -> SessionRecord:
        pass

    @abc.abstractmethod
    def store_session(
        self, recipient_id: str, device_id: int, session_record: SessionRecord
    ) -> None:
        pass

    @abc.abstractmethod
    def contains_session(self, recipient_id: str, device_id: int) -> bool:
        pass

    @abc.abstractmethod
    def get_inactive_sessions_keys(self, recipient_id: str) -> list[IdentityKey]:
        pass

    @abc.abstractmethod
    def delete_session(self, recipient_id: str, device_id: int) -> None:
        pass

    @abc.abstractmethod
    def delete_all_sessions(self, recipient_id: str) -> None:
        pass

    # Others

    @abc.abstractmethod
    def get_active_device_tuples(self) -> list[tuple[str, int]]:
        pass

    @abc.abstractmethod
    def get_trust_for_identity(
        self, recipient_id: str, identity_key: IdentityKey
    ) -> Optional[OMEMOTrust]:
        pass

    @abc.abstractmethod
    def get_unacknowledged_count(self, recipient_id: str, device_id: int) -> int:
        pass

    @abc.abstractmethod
    def set_active_state(self, address: str, devicelist: list[int]) -> None:
        pass

    @abc.abstractmethod
    def set_inactive(self, address: str, device_id: int) -> None:
        pass

    @abc.abstractmethod
    def is_trusted(self, recipient_id: str, device_id: int) -> bool:
        pass

    @abc.abstractmethod
    def needs_init(self) -> bool:
        pass



from __future__ import annotations

from ..ecc.curve import Curve
from ..ecc.djbec import CurvePublicKey
from ..ecc.eckeypair import ECKeyPair
from ..kdf import hkdf
from ..kdf.derivedrootsecrets import DerivedRootSecrets
from .chainkey import ChainKey


class RootKey:
    def __init__(self, session_version: int, key: bytes) -> None:
        self._session_version = session_version
        self._key = key

    def get_key_bytes(self) -> bytes:
        return self._key

    def create_chain(
        self,
        ec_public_key_their_ratchet_key: CurvePublicKey,
        ec_key_pair_our_ratchet_key: ECKeyPair,
    ) -> tuple[RootKey, ChainKey]:
        if self._session_version == 3:
            domain_separator = "WhisperRatchet"
        elif self._session_version == 4:
            domain_separator = "OMEMO Root Chain"
        else:
            raise ValueError("Invalid session version: %s", self._session_version)

        shared_secret = Curve.calculate_agreement(
            ec_public_key_their_ratchet_key,
            ec_key_pair_our_ratchet_key.get_private_key(),
        )

        derived_secret_bytes = hkdf.derive(
            input_key_material=shared_secret,
            length=DerivedRootSecrets.SIZE,
            salt=self._key,
            info=domain_separator.encode(),
        )

        derived_secrets = DerivedRootSecrets(derived_secret_bytes)
        new_root_key = RootKey(self._session_version, derived_secrets.get_root_key())
        new_chain_key = ChainKey(
            self._session_version, derived_secrets.get_chain_key(), 0
        )
        return (new_root_key, new_chain_key)



from __future__ import annotations

import random
import time

from ..const import MAX_INT
from ..ecc.curve import Curve
from ..identitykey import IdentityKey
from ..identitykeypair import IdentityKeyPair
from ..state.prekeyrecord import PreKeyRecord
from ..state.signedprekeyrecord import SignedPreKeyRecord


class KeyHelper:
    @staticmethod
    def generate_identity_key_pair() -> IdentityKeyPair:
        """
        Generate an identity key pair. Clients should only do this once,
        at install time.
        @return the generated IdentityKeyPair.
        """
        key_pair = Curve.generate_key_pair()
        public_key = IdentityKey(key_pair.get_public_key())
        identity_key_pair = IdentityKeyPair.new(public_key, key_pair.get_private_key())
        return identity_key_pair

    @staticmethod
    def get_random_int() -> int:
        return random.randint(1, MAX_INT)

    @staticmethod
    def generate_pre_keys(start: int, count: int) -> list[PreKeyRecord]:
        """
        Generate a list of PreKeys.  Clients should do this at install time, and
        subsequently any time the list of PreKeys stored on the server runs low.

        PreKey IDs are shorts, so they will eventually be repeated.
        Clients should store PreKeys in a circular buffer, so that they are
        repeated as infrequently as possible.
        """
        results: list[PreKeyRecord] = []
        start -= 1
        for i in range(0, count):
            pre_key_id = ((start + i) % MAX_INT) + 1
            results.append(PreKeyRecord.new(pre_key_id, Curve.generate_key_pair()))

        return results

    @staticmethod
    def generate_signed_pre_key(
        identity_key_pair: IdentityKeyPair, signed_pre_key_id: int
    ) -> SignedPreKeyRecord:
        key_pair = Curve.generate_key_pair()
        signature = Curve.calculate_signature(
            identity_key_pair.get_private_key(), key_pair.get_public_key().serialize()
        )

        spk = SignedPreKeyRecord.new(
            signed_pre_key_id, int(round(time.time() * 1000)), key_pair, signature
        )

        return spk

    @staticmethod
    def get_next_signed_pre_key_id(current_id: int) -> int:
        return current_id % MAX_INT + 1



import abc


class ECPublicKey(abc.ABC):
    @abc.abstractmethod
    def serialize(self) -> bytes:
        pass


class ECPrivateKey(abc.ABC):
    @abc.abstractmethod
    def serialize(self) -> bytes:
        pass



from __future__ import annotations

from typing import cast

import google.protobuf.message

from .ecc.curve import Curve
from .ecc.djbec import DjbECPrivateKey
from .identitykey import IdentityKey
from .state import storage_pb2


class IdentityKeyPair:
    def __init__(self, structure: IdentityKeyPairStructureProto) -> None:
        self._structure = structure

    @classmethod
    def new(
        cls,
        identity_key_public_key: IdentityKey,
        ec_private_key: DjbECPrivateKey,
    ) -> IdentityKeyPair:
        structure = cast(
            IdentityKeyPairStructureProto,
            storage_pb2.IdentityKeyPairStructure(),  # pyright: ignore
        )

        structure.publicKey = identity_key_public_key.serialize()
        structure.privateKey = ec_private_key.serialize()

        return cls(structure)

    @classmethod
    def from_bytes(cls, serialized: bytes) -> IdentityKeyPair:
        structure = cast(
            IdentityKeyPairStructureProto,
            storage_pb2.IdentityKeyPairStructure(),  # pyright: ignore
        )
        structure.ParseFromString(serialized)
        return cls(structure)

    def get_public_key(self) -> IdentityKey:
        return IdentityKey(self._structure.publicKey)

    def get_private_key(self) -> DjbECPrivateKey:
        return Curve.decode_private_point(self._structure.privateKey)

    def serialize(self) -> bytes:
        return self._structure.SerializeToString()


class IdentityKeyPairStructureProto(google.protobuf.message.Message):
    publicKey: bytes
    privateKey: bytes



from __future__ import annotations

import hashlib
import hmac

from ..kdf import hkdf
from ..kdf.derivedmessagesecrets import DerivedMessageSecrets
from ..kdf.messagekeys import MessageKeys

MESSAGE_KEY_SEED = bytes([0x01])
CHAIN_KEY_SEED = bytes([0x02])


class ChainKey:
    def __init__(self, session_version: int, key: bytes, index: int) -> None:
        self._session_version = session_version
        self._key = key
        self._index = index

    def get_key(self) -> bytes:
        return self._key

    def get_index(self) -> int:
        return self._index

    def get_next_chain_key(self) -> ChainKey:
        nextKey = self.get_base_material(CHAIN_KEY_SEED)
        return ChainKey(self._session_version, nextKey, self._index + 1)

    def get_message_keys(self) -> MessageKeys:
        if self._session_version == 3:
            domain_separator = "WhisperMessageKeys"
        elif self._session_version == 4:
            domain_separator = "OMEMO Message Key Material"
        else:
            raise ValueError("Invalid session version: %s", self._session_version)

        input_key_material = self.get_base_material(MESSAGE_KEY_SEED)
        key_material_bytes = hkdf.derive(
            input_key_material=input_key_material,
            length=DerivedMessageSecrets.SIZE,
            salt=bytes(32),
            info=domain_separator.encode(),
        )
        key_material = DerivedMessageSecrets(key_material_bytes)

        return MessageKeys(
            key_material.get_cipher_key(),
            key_material.get_mac_key(),
            key_material.get_iv(),
            self._index,
        )

    def get_base_material(self, seedBytes: bytes) -> bytes:
        mac = hmac.new(bytes(self._key), digestmod=hashlib.sha256)
        mac.update(bytes(seedBytes))
        return mac.digest()



from __future__ import annotations

from typing import Union

import os

from .. import curve
from .djbec import CurvePublicKey
from .djbec import DjbECPrivateKey
from .djbec import EdPublicKey
from .eckeypair import ECKeyPair


class Curve:
    @staticmethod
    def _generate_private_key() -> bytes:
        rand = os.urandom(32)
        return curve.generate_private_key(rand)

    @staticmethod
    def _generate_public_key(private_key: bytes) -> bytes:
        return curve.generate_public_key(private_key)

    @staticmethod
    def generate_key_pair() -> ECKeyPair:
        private_key = Curve._generate_private_key()
        public_key = Curve._generate_public_key(private_key)
        return ECKeyPair(CurvePublicKey(public_key), DjbECPrivateKey(private_key))

    @staticmethod
    def decode_point(bytes_: bytes) -> CurvePublicKey:
        return CurvePublicKey.from_bytes(bytes_)

    @staticmethod
    def decode_private_point(_bytes: bytes) -> DjbECPrivateKey:
        return DjbECPrivateKey(_bytes)

    @staticmethod
    def calculate_agreement(
        public_key: CurvePublicKey, private_key: DjbECPrivateKey
    ) -> bytes:
        return curve.calculate_agreement(
            private_key.get_private_key(), public_key.get_bytes()
        )

    @staticmethod
    def verify_signature(
        ec_public_signing_key: Union[CurvePublicKey, EdPublicKey],
        message: bytes,
        signature: bytes,
    ) -> bool:
        if isinstance(ec_public_signing_key, CurvePublicKey):
            result = curve.verify_signature_curve(
                ec_public_signing_key.get_bytes(), message, signature
            )
        else:
            result = curve.verify_signature_ed(
                ec_public_signing_key.get_bytes(), message, signature
            )
        return result == 0

    @staticmethod
    def calculate_signature(
        private_signing_key: DjbECPrivateKey, message: bytes
    ) -> bytes:
        rand = os.urandom(64)
        return curve.calculate_signature(
            rand, private_signing_key.get_private_key(), message
        )



from __future__ import annotations

from typing import NamedTuple

import logging
import os

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers import modes
from cryptography.hazmat.primitives.ciphers.modes import GCM
from cryptography.hazmat.primitives.hmac import HMAC
from cryptography.hazmat.primitives.padding import PKCS7

log = logging.getLogger(__name__)

IV_SIZE = 12


class EncryptionResult(NamedTuple):
    payload: bytes
    key: bytes
    iv: bytes


def _decrypt(key: bytes, iv: bytes, tag: bytes, data: bytes) -> bytes:
    decryptor = Cipher(
        algorithms.AES(key), GCM(iv, tag=tag), backend=default_backend()
    ).decryptor()
    return decryptor.update(data) + decryptor.finalize()


def aes_decrypt(_key: bytes, iv: bytes, payload: bytes) -> str:
    if len(_key) >= 32:
        # XEP-0384
        log.debug("XEP Compliant Key/Tag")
        data = payload
        key = _key[:16]
        tag = _key[16:]
    else:
        # Legacy
        log.debug("Legacy Key/Tag")
        data = payload[:-16]
        key = _key
        tag = payload[-16:]

    return _decrypt(key, iv, tag, data).decode()


def aes_decrypt_file(key: bytes, iv: bytes, payload: bytes) -> bytes:
    data = payload[:-16]
    tag = payload[-16:]
    return _decrypt(key, iv, tag, data)


def aes_gcm_encrypt(key: bytes, iv: bytes, plaintext: bytes) -> tuple[bytes, bytes]:
    encryptor = Cipher(
        algorithms.AES(key), GCM(iv), backend=default_backend()
    ).encryptor()

    ciphertext = encryptor.update(plaintext) + encryptor.finalize()
    return ciphertext, encryptor.tag


def _encrypt(
    data: str | bytes, key_size: int, iv_size: int = IV_SIZE
) -> tuple[bytes, bytes, bytes, bytes]:
    if isinstance(data, str):
        data = data.encode()
    key = os.urandom(key_size)
    iv = os.urandom(iv_size)
    encryptor = Cipher(
        algorithms.AES(key), GCM(iv), backend=default_backend()
    ).encryptor()

    payload = encryptor.update(data) + encryptor.finalize()
    return key, iv, encryptor.tag, payload


def aes_encrypt(plaintext: str) -> EncryptionResult:
    key, iv, tag, payload = _encrypt(plaintext, 16)
    key += tag
    return EncryptionResult(payload=payload, key=key, iv=iv)


def aes_encrypt_file(data: bytes) -> EncryptionResult:
    (
        key,
        iv,
        tag,
        payload,
    ) = _encrypt(data, 32)
    payload += tag
    return EncryptionResult(payload=payload, key=key, iv=iv)


def get_new_key() -> bytes:
    return os.urandom(16)


def get_new_iv() -> bytes:
    return os.urandom(IV_SIZE)


def aes_cbc_encrypt(key: bytes, iv: bytes, plaintext: bytes) -> bytes:
    padder = PKCS7(128).padder()
    padded_plaintext = padder.update(plaintext) + padder.finalize()

    encryptor = Cipher(
        algorithms.AES(key), modes.CBC(iv), backend=default_backend()
    ).encryptor()

    return encryptor.update(padded_plaintext) + encryptor.finalize()


def aes_cbc_decrypt(key: bytes, iv: bytes, ciphertext: bytes) -> bytes:
    decryptor = Cipher(
        algorithms.AES(key), modes.CBC(iv), backend=default_backend()
    ).decryptor()
    padded_plaintext = decryptor.update(ciphertext) + decryptor.finalize()

    unpadder = PKCS7(128).unpadder()
    return unpadder.update(padded_plaintext) + unpadder.finalize()


def hmac_calculate(key: bytes, data: bytes) -> bytes:
    hmac = HMAC(key, hashes.SHA256(), backend=default_backend())
    hmac.update(data)
    return hmac.finalize()



from __future__ import annotations

from typing import cast

import hashlib
import hmac

import google.protobuf.message

from ..ecc.curve import Curve
from ..ecc.djbec import CurvePublicKey
from ..exceptions import InvalidKeyException
from ..exceptions import InvalidMessageException
from ..identitykey import IdentityKey
from ..util.byteutil import ByteUtil
from . import whisper_pb2
from .ciphertextmessage import CiphertextMessage

MAC_LENGTH = 8


class WhisperMessage(CiphertextMessage):
    def __init__(
        self,
        serialized: bytes,
        sender_ratchet_key: CurvePublicKey,
        counter: int,
        previous_counter: int,
        ciphertext: bytes,
    ) -> None:
        self._serialized = serialized
        self._sender_ratchet_key = sender_ratchet_key
        self._counter = counter
        self._previous_counter = previous_counter
        self._ciphertext = ciphertext

    @classmethod
    def new(
        cls,
        message_version: int,
        mac_key: bytes,
        ec_public_key_sender_ratchet_key: CurvePublicKey,
        counter: int,
        previous_counter: int,
        ciphertext: bytes,
        sender_identity_key: IdentityKey,
        receiver_identity_key: IdentityKey,
    ) -> WhisperMessage:
        version = ByteUtil.ints_to_byte_high_and_low(message_version, message_version)

        message = cast(
            WhisperMessageProto, whisper_pb2.WhisperMessage()  # pyright: ignore
        )
        message.ratchetKey = ec_public_key_sender_ratchet_key.serialize()
        message.counter = counter
        message.previousCounter = previous_counter
        message.ciphertext = ciphertext
        message = message.SerializeToString()

        mac = cls.get_mac(
            sender_identity_key,
            receiver_identity_key,
            mac_key,
            ByteUtil.combine(version, message),
        )

        serialized = ByteUtil.combine(version, message, mac)

        return cls(
            serialized,
            ec_public_key_sender_ratchet_key,
            counter,
            previous_counter,
            ciphertext,
        )

    @classmethod
    def from_bytes(cls, serialized: bytes) -> WhisperMessage:
        message_parts = ByteUtil.split(
            serialized,
            1,
            len(serialized) - 1 - MAC_LENGTH,
            MAC_LENGTH,
        )

        version = ByteUtil.high_bits_to_int(message_parts[0][0])
        message = message_parts[1]

        if version != 3:
            raise InvalidMessageException("Unknown version: %s" % version)

        whisper_message = cast(
            WhisperMessageProto, whisper_pb2.WhisperMessage()  # pyright: ignore
        )
        whisper_message.ParseFromString(message)

        if not whisper_message.ciphertext or not whisper_message.ratchetKey:
            raise InvalidMessageException("Incomplete message")

        try:
            sender_ratchet_key = Curve.decode_point(whisper_message.ratchetKey)
            assert isinstance(sender_ratchet_key, CurvePublicKey)
        except InvalidKeyException as e:
            raise InvalidMessageException(str(e))

        return WhisperMessage(
            serialized,
            sender_ratchet_key,
            whisper_message.counter,
            whisper_message.previousCounter,
            whisper_message.ciphertext,
        )

    def get_sender_ratchet_key(self) -> CurvePublicKey:
        return self._sender_ratchet_key

    def get_message_version(self) -> int:
        return 3

    def get_counter(self) -> int:
        return self._counter

    def get_body(self) -> bytes:
        return self._ciphertext

    def verify_mac(
        self,
        sender_identity_key: IdentityKey,
        receiver_identity_key: IdentityKey,
        mac_key: bytes,
    ) -> None:
        parts = ByteUtil.split(
            self._serialized, len(self._serialized) - MAC_LENGTH, MAC_LENGTH
        )
        our_mac = self.get_mac(
            sender_identity_key, receiver_identity_key, mac_key, parts[0]
        )
        their_mac = parts[1]

        if our_mac != their_mac:
            raise InvalidMessageException("Bad Mac!")

    @classmethod
    def get_mac(
        cls,
        sender_identity_key: IdentityKey,
        receiver_identity_key: IdentityKey,
        mac_key: bytes,
        serialized: bytes,
    ) -> bytes:
        mac = hmac.new(mac_key, digestmod=hashlib.sha256)
        mac.update(sender_identity_key.get_public_key().serialize())
        mac.update(receiver_identity_key.get_public_key().serialize())
        mac.update(serialized)
        full_mac = mac.digest()
        return ByteUtil.trim(full_mac, MAC_LENGTH)

    def serialize(self) -> bytes:
        return self._serialized

    def get_type(self) -> int:
        return CiphertextMessage.WHISPER_TYPE


class WhisperMessageProto(google.protobuf.message.Message):
    ratchetKey: bytes
    counter: int
    previousCounter: int
    ciphertext: bytes



from __future__ import annotations

from ..ecc.curve import Curve
from ..ecc.djbec import CurvePublicKey
from ..kdf import hkdf
from ..state.sessionstate import SessionState
from ..util.byteutil import ByteUtil
from .aliceparameters import AliceParameters
from .bobparameters import BobParameters
from .chainkey import ChainKey
from .rootkey import RootKey


class RatchetingSession:
    @staticmethod
    def initialize_session_as_alice(
        session_state: SessionState,
        session_version: int,
        parameters: AliceParameters,
    ) -> None:
        session_state.set_session_version(session_version)
        session_state.set_remote_identity_key(parameters.get_their_identity_key())
        session_state.set_local_identity_key(
            parameters.get_our_identity_key().get_public_key()
        )

        sending_ratchet_key = Curve.generate_key_pair()
        secrets = bytearray()

        secrets.extend(RatchetingSession.get_discontinuity_bytes())

        secrets.extend(
            Curve.calculate_agreement(
                parameters.get_their_signed_pre_key(),
                parameters.get_our_identity_key().get_private_key(),
            )
        )
        secrets.extend(
            Curve.calculate_agreement(
                parameters.get_their_identity_key().get_public_key(),
                parameters.get_our_base_key().get_private_key(),
            )
        )
        secrets.extend(
            Curve.calculate_agreement(
                parameters.get_their_signed_pre_key(),
                parameters.get_our_base_key().get_private_key(),
            )
        )

        secrets.extend(
            Curve.calculate_agreement(
                parameters.get_their_one_time_pre_key(),
                parameters.get_our_base_key().get_private_key(),
            )
        )

        derived_keys = RatchetingSession.calculate_derived_keys(
            session_version, bytes(secrets)
        )
        sending_chain = derived_keys.get_root_key().create_chain(
            parameters.get_their_ratchet_key(), sending_ratchet_key
        )

        session_state.add_receiver_chain(
            parameters.get_their_ratchet_key(), derived_keys.get_chain_key()
        )
        session_state.set_sender_chain(sending_ratchet_key, sending_chain[1])
        session_state.set_root_key(sending_chain[0])

    @staticmethod
    def initialize_session_as_bob(
        session_state: SessionState,
        session_version: int,
        parameters: BobParameters,
    ) -> None:
        session_state.set_session_version(session_version)
        session_state.set_remote_identity_key(parameters.get_their_identity_key())
        session_state.set_local_identity_key(
            parameters.get_our_identity_key().get_public_key()
        )

        secrets = bytearray()

        secrets.extend(RatchetingSession.get_discontinuity_bytes())

        secrets.extend(
            Curve.calculate_agreement(
                parameters.get_their_identity_key().get_public_key(),
                parameters.get_our_signed_pre_key().get_private_key(),
            )
        )

        secrets.extend(
            Curve.calculate_agreement(
                parameters.get_their_base_key(),
                parameters.get_our_identity_key().get_private_key(),
            )
        )
        secrets.extend(
            Curve.calculate_agreement(
                parameters.get_their_base_key(),
                parameters.get_our_signed_pre_key().get_private_key(),
            )
        )

        our_one_time_pre_key = parameters.get_our_one_time_pre_key()
        secrets.extend(
            Curve.calculate_agreement(
                parameters.get_their_base_key(),
                our_one_time_pre_key.get_private_key(),
            )
        )

        derived_keys = RatchetingSession.calculate_derived_keys(
            session_version, bytes(secrets)
        )
        session_state.set_sender_chain(
            parameters.get_our_ratchet_key(), derived_keys.get_chain_key()
        )
        session_state.set_root_key(derived_keys.get_root_key())

    @staticmethod
    def get_discontinuity_bytes() -> bytearray:
        return bytearray([0xFF] * 32)

    @staticmethod
    def calculate_derived_keys(
        session_version: int, master_secret: bytes
    ) -> DerivedKeys:
        if session_version <= 3:
            domain_separator = "WhisperText"
        else:
            domain_separator = "OMEMO Payload"

        derived_secret_bytes = hkdf.derive(
            input_key_material=master_secret,
            length=64,
            salt=bytes(32),
            info=domain_separator.encode(),
        )
        derived_secrets = ByteUtil.split(derived_secret_bytes, 32, 32)

        return RatchetingSession.DerivedKeys(
            RootKey(session_version, derived_secrets[0]),
            ChainKey(session_version, derived_secrets[1], 0),
        )

    @staticmethod
    def is_alice(our_key: CurvePublicKey, their_key: CurvePublicKey) -> bool:
        return our_key < their_key

    class DerivedKeys:
        def __init__(self, root_key: RootKey, chain_key: ChainKey):
            self._root_key = root_key
            self._chain_key = chain_key

        def get_root_key(self):
            return self._root_key

        def get_chain_key(self):
            return self._chain_key



from __future__ import annotations

from typing import Any
from typing import Optional

import logging
import time
from collections import defaultdict

from . import exceptions
from .aes import aes_decrypt
from .aes import aes_encrypt
from .aes import get_new_iv
from .aes import get_new_key
from .const import NS_OMEMO_TMP
from .const import OMEMOTrust
from .identitykey import IdentityKey
from .identitykeypair import IdentityKeyPair
from .observable import Observable
from .protocol.prekeywhispermessage import PreKeyWhisperMessage
from .protocol.whispermessage import WhisperMessage
from .sessionbuilder import SessionBuilder
from .sessioncipher import SessionCipher
from .state.prekeybundle import PreKeyBundle
from .state.store import Store
from .structs import IdentityInfo
from .structs import OMEMOBundle
from .structs import OMEMOBundleProto
from .structs import OMEMOConfig
from .structs import OMEMOMessage
from .structs import OMEMOMessageProto
from .util.keyhelper import KeyHelper
from .util.logging import SessionContextAdapter

log = logging.getLogger(__name__)


class OMEMOSessionManager(Observable):
    def __init__(
        self,
        our_address: str,
        storage: Store,
        config: OMEMOConfig,
        log_context: str | None = None,
    ) -> None:
        Observable.__init__(self)
        self._log = SessionContextAdapter(log, {"context": log_context})

        self._device_store: dict[str, set[int]] = defaultdict(set)
        self._group_member_store: dict[str, set[str]] = defaultdict(set)
        self._session_ciphers: dict[str, dict[int, SessionCipher]] = defaultdict(dict)

        self._our_address = our_address
        self._storage = storage
        self._config = config

        if self._storage.needs_init():
            self._log.info("Generating OMEMO keys")
            self._our_device_id = self._generate_keys()
        else:
            self._our_device_id = self._storage.get_our_device_id()

        self.add_device(self._our_address, self._our_device_id)
        self._log.info("Our device id: %s", self._our_device_id)

        self._log.info("%s PreKeys available", self._storage.get_pre_key_count())

        for address, device in self._storage.get_active_device_tuples():
            self._log.info("Load device from storage: %s - %s", address, device)
            self.add_device(address, device)

    def _generate_keys(self) -> int:
        identity_key_pair = KeyHelper.generate_identity_key_pair()
        our_device_id = KeyHelper.get_random_int()
        self._storage.set_our_identity(our_device_id, identity_key_pair)

        signed_pre_key = KeyHelper.generate_signed_pre_key(
            identity_key_pair, KeyHelper.get_random_int()
        )
        self._storage.store_signed_pre_key(signed_pre_key.get_id(), signed_pre_key)

        self._generate_pre_keys(self._config.default_prekey_amount)

        return our_device_id

    def _generate_pre_keys(self, count: int) -> None:
        prekey_id = self._storage.get_current_pre_key_id() or 0
        pre_keys = KeyHelper.generate_pre_keys(prekey_id + 1, count)
        for pre_key in pre_keys:
            self._storage.store_pre_key(pre_key.get_id(), pre_key)

    def build_session(self, address: str, bundle: OMEMOBundleProto) -> None:
        session = SessionBuilder(self._storage, address, bundle.device_id)
        prekey_bundle = PreKeyBundle.from_proto(bundle)

        session.process_pre_key_bundle(prekey_bundle)

    def delete_session(
        self, address: str, device_id: int, *, delete_identity: bool
    ) -> None:
        if delete_identity:
            record = self._storage.load_session(address, device_id)
            identity_key = record.get_session_state().get_remote_identity_key()
            self._storage.delete_identity(address, identity_key)
            self.remove_device(address, device_id)

        self._storage.delete_session(address, device_id)

    def get_our_identity(self) -> tuple[int, IdentityKey]:
        our_device_id = self._storage.get_our_device_id()
        identity_key = self._storage.get_identity_key_pair().get_public_key()
        return our_device_id, identity_key

    def get_our_fingerprint(self, *, formatted: bool = False) -> str:
        identity_key = self._storage.get_identity_key_pair().get_public_key()
        return identity_key.get_fingerprint(formatted=formatted)

    def get_bundle(self, namespace: str) -> OMEMOBundle:
        self._check_pre_key_count()

        bundle: dict[str, Any] = {"otpks": []}
        for k in self._storage.load_pending_pre_keys():
            key = k.get_key_pair().get_public_key().serialize()
            bundle["otpks"].append({"key": key, "id": k.get_id()})

        ik_pair = self._storage.get_identity_key_pair()
        bundle["ik"] = ik_pair.get_public_key().serialize()

        self._cycle_signed_pre_key(ik_pair)

        spk = self._storage.load_signed_pre_key(
            self._storage.get_current_signed_pre_key_id()
        )
        bundle["spk_signature"] = spk.get_signature()
        bundle["spk"] = {
            "key": spk.get_key_pair().get_public_key().serialize(),
            "id": spk.get_id(),
        }

        device_id = self.get_our_device()

        return OMEMOBundle(**bundle, device_id=device_id, namespace=namespace)

    def decrypt_message(
        self, omemo_message: OMEMOMessageProto, address: str
    ) -> tuple[str, str, OMEMOTrust]:
        if omemo_message.sid == self.get_our_device():
            self._log.info("Received previously sent message by us")
            raise exceptions.SelfMessage

        try:
            encrypted_key, prekey = omemo_message.keys[self.get_our_device()]
        except KeyError:
            self._log.info("Received message not for our device")
            raise exceptions.MessageNotForDevice

        try:
            if prekey:
                key, fingerprint, trust = self._process_pre_key_message(
                    address, omemo_message.sid, encrypted_key
                )
            else:
                key, fingerprint, trust = self._process_message(
                    address, omemo_message.sid, encrypted_key
                )

        except exceptions.DuplicateMessageException:
            self._log.info("Received duplicated message")
            raise exceptions.DuplicateMessage

        except Exception as error:
            self._log.warning(error)
            raise exceptions.DecryptionFailed

        if omemo_message.payload is None:
            self._log.debug("Decrypted Key Exchange Message")
            raise exceptions.KeyExchangeMessage

        try:
            result = aes_decrypt(key, omemo_message.iv, omemo_message.payload)
        except Exception as error:
            self._log.warning(error)
            raise exceptions.DecryptionFailed

        self._log.debug("Decrypted Message => %s", result)
        return result, fingerprint, trust

    def _get_whisper_message(
        self, address: str, device: int, key: bytes
    ) -> tuple[bytes, bool]:
        session_cipher = SessionCipher(self._storage, address, device)
        cipher_key = session_cipher.encrypt(key)
        prekey = isinstance(cipher_key, PreKeyWhisperMessage)
        return cipher_key.serialize(), prekey

    def encrypt(
        self, address: str, plaintext: str, *, groupchat: bool
    ) -> Optional[OMEMOMessage]:
        try:
            devices_for_encryption = self._get_devices_for_enc(
                address, groupchat=groupchat
            )
        except exceptions.NoDevicesFound:
            self._log.warning("No devices for encryption found for: %s", address)
            return

        result = aes_encrypt(plaintext)
        whisper_messages: dict[str, dict[int, tuple[bytes, bool]]] = defaultdict(dict)

        for address_, device in devices_for_encryption:
            count = self._storage.get_unacknowledged_count(address_, device)
            if count >= self._config.unacknowledged_count:
                self._log.warning(
                    "Set device inactive %s because of %s unacknowledged messages",
                    device,
                    count,
                )
                self.remove_device(address_, device)

            try:
                whisper_messages[address_][device] = self._get_whisper_message(
                    address_, device, result.key
                )
            except Exception:
                self._log.exception("Failed to encrypt")
                continue

        if not whisper_messages:
            self._log.error("No encrypted messages")
            return

        encrypted_keys: dict[int, tuple[bytes, bool]] = {}
        for address_ in whisper_messages:
            encrypted_keys.update(whisper_messages[address_])

        self._log.debug("Finished encrypting message")
        return OMEMOMessage(
            sid=self.get_our_device(),
            keys=encrypted_keys,
            iv=result.iv,
            payload=result.payload,
        )

    def encrypt_key_transport(
        self, address: str, devices: list[int]
    ) -> Optional[OMEMOMessage]:
        whisper_messages: dict[str, dict[int, tuple[bytes, bool]]] = defaultdict(dict)
        for device in devices:
            try:
                whisper_messages[address][device] = self._get_whisper_message(
                    address, device, get_new_key()
                )
            except Exception:
                self._log.exception("Failed to encrypt")
                continue

        if not whisper_messages[address]:
            self._log.error("Encrypted keys empty")
            return

        self._log.debug("Finished Key Transport message")
        return OMEMOMessage(
            sid=self.get_our_device(),
            keys=whisper_messages[address],
            iv=get_new_iv(),
            payload=None,
        )

    def set_trust(
        self, address: str, identity_key: IdentityKey, trust: OMEMOTrust
    ) -> None:
        self._storage.set_trust(address, identity_key, trust)

    def get_identity_infos(
        self,
        address: str,
        *,
        only_active: bool = False,
        trust: list[OMEMOTrust] | OMEMOTrust | None = None,
    ) -> list[IdentityInfo]:
        if self._is_group_address(address):
            members = list(self.get_group_members(address))
            identity_infos = self._storage.get_identity_infos(members)
        else:
            identity_infos = self._storage.get_identity_infos(address)

        if only_active:
            identity_infos = list(filter(lambda info: info.active, identity_infos))

        if trust is not None:
            if not isinstance(trust, list):
                trust = [trust]
            identity_infos = list(
                filter(lambda info: info.trust in trust, identity_infos)
            )

        return identity_infos

    def get_devices_without_sessions(self, address: str) -> list[int]:
        known_devices = self.get_devices(address, without_self=True)
        missing_devices = [
            dev
            for dev in known_devices
            if not self._storage.contains_session(address, dev)
        ]
        if missing_devices:
            self._log.info(
                "Missing device sessions for %s: %s", address, missing_devices
            )
        return missing_devices

    def _process_pre_key_message(
        self, address: str, device: int, key: bytes
    ) -> tuple[bytes, str, OMEMOTrust]:
        self._log.info("Process pre key message from %s", address)
        pre_key_message = PreKeyWhisperMessage.from_bytes(key)
        if not pre_key_message.get_pre_key_id():
            raise Exception("Received Pre Key Message without PreKey => %s" % address)

        session_cipher = SessionCipher(self._storage, address, device)
        key = session_cipher.decrypt_pkmsg(pre_key_message)

        identity_key = pre_key_message.get_identity_key()
        trust = self._get_trust_from_identity_key(address, identity_key)
        fingerprint = identity_key.get_fingerprint()

        self._storage.set_identity_last_seen(address, identity_key)

        self.add_device(address, device)
        self._notify("republish-bundle", self.get_bundle(NS_OMEMO_TMP))
        return key, fingerprint, trust

    def _process_message(
        self, address: str, device: int, key: bytes
    ) -> tuple[bytes, str, OMEMOTrust]:
        self._log.info("Process message from %s", address)
        message = WhisperMessage.from_bytes(key)

        session_cipher = SessionCipher(self._storage, address, device)
        key = session_cipher.decrypt_msg(message)

        identity_key = self._get_identity_key_from_device(address, device)
        assert identity_key is not None
        trust = self._get_trust_from_identity_key(address, identity_key)
        fingerprint = identity_key.get_fingerprint()

        self._storage.set_identity_last_seen(address, identity_key)

        self.add_device(address, device)

        return key, fingerprint, trust

    def _get_identity_key_from_device(
        self, address: str, device: int
    ) -> Optional[IdentityKey]:
        session_record = self._storage.load_session(address, device)
        return session_record.get_session_state().get_remote_identity_key()

    def _get_trust_from_identity_key(
        self, address: str, identity_key: IdentityKey
    ) -> OMEMOTrust:
        trust = self._storage.get_trust_for_identity(address, identity_key)
        return OMEMOTrust(trust) if trust is not None else OMEMOTrust.UNDECIDED

    def _check_pre_key_count(self) -> None:
        # Check if enough PreKeys are available
        pre_key_count = self._storage.get_pre_key_count()
        if pre_key_count < self._config.min_prekey_amount:
            missing_count = self._config.default_prekey_amount - pre_key_count
            self._generate_pre_keys(missing_count)
            self._log.info("%s PreKeys created", missing_count)

    def _cycle_signed_pre_key(self, ik_pair: IdentityKeyPair) -> None:
        # Publish every spk_cycle_seconds a new SignedPreKey
        # Delete all existing SignedPreKeys that are older
        # then spk_archive_seconds

        # if spk_cycle_seconds is reached, generate a new SignedPreKey
        now = int(time.time())
        current_spk_id = self._storage.get_current_signed_pre_key_id()
        timestamp = self._storage.get_signed_pre_key_timestamp(current_spk_id)

        if int(timestamp) < now - self._config.spk_cycle_seconds:
            spk = KeyHelper.generate_signed_pre_key(
                ik_pair, KeyHelper.get_next_signed_pre_key_id(current_spk_id)
            )
            self._storage.store_signed_pre_key(spk.get_id(), spk)
            self._log.debug("Cycled SignedPreKey")

        # Delete all SignedPreKeys that are older than spk_archive_seconds
        timestamp = now - self._config.spk_archive_seconds
        self._storage.remove_old_signed_pre_keys(timestamp)

    def update_devicelist(self, address: str, devicelist: list[int]) -> None:
        for device in list(devicelist):
            if device == self.get_our_device():
                continue
            count = self._storage.get_unacknowledged_count(address, device)
            if count > self._config.unacknowledged_count:
                self._log.warning(
                    "Ignore device because of %s unacknowledged messages: %s %s",
                    count,
                    address,
                    device,
                )
                devicelist.remove(device)

        self._device_store[address] = set(devicelist)
        self._log.info("Saved devices for %s", address)
        self._storage.set_active_state(address, devicelist)

    def _is_group_address(self, address: str) -> bool:
        return address in self._group_member_store

    def clear_group_members(self, group_address: str) -> None:
        self._group_member_store[group_address] = set()

    def set_group_members(self, group_address: str, group_members: set[str]) -> None:
        self._group_member_store[group_address] = group_members

    def add_group_member(self, group_address: str, address: str) -> None:
        self._log.info("Saved group member %s %s", group_address, address)
        self._group_member_store[group_address].add(address)

    def remove_group_member(self, group_address: str, address: str) -> None:
        self._log.info("Removed group member %s %s", group_address, address)
        self._group_member_store[group_address].discard(address)

    def get_group_members(
        self, group_address: str, without_self: bool = True
    ) -> set[str]:
        members = set(self._group_member_store[group_address])
        if without_self:
            members.discard(self._our_address)
        return members

    def add_device(self, address: str, device: int) -> None:
        self._device_store[address].add(device)

    def remove_device(self, address: str, device: int) -> None:
        self._device_store[address].discard(device)
        self._storage.set_inactive(address, device)

    def get_devices(self, address: str, without_self: bool = False) -> set[int]:
        devices = set(self._device_store[address])
        if without_self:
            devices.discard(self.get_our_device())
        return devices

    def _get_devices_for_enc(
        self, address: str, *, groupchat: bool
    ) -> set[tuple[str, int]]:
        devices_for_encryption: list[tuple[str, int]] = []

        if groupchat:
            devices_for_encryption = self._get_devices_for_group_encryption(address)
        else:
            devices_for_encryption = self._get_devices_for_encryption(address)

        if not devices_for_encryption:
            raise exceptions.NoDevicesFound

        devices_for_encryption += self._get_our_devices_for_encryption()
        return set(devices_for_encryption)

    def _get_devices_for_group_encryption(self, address: str) -> list[tuple[str, int]]:
        devices_for_encryption: list[tuple[str, int]] = []
        for address_ in self._group_member_store[address]:
            devices_for_encryption += self._get_devices_for_encryption(address_)
        return devices_for_encryption

    def _get_our_devices_for_encryption(self) -> list[tuple[str, int]]:
        devices_for_encryption: list[tuple[str, int]] = []
        our_devices = self.get_devices(self._our_address, without_self=True)
        for device in our_devices:
            if self._storage.is_trusted(self._our_address, device):
                devices_for_encryption.append((self._our_address, device))
        return devices_for_encryption

    def _get_devices_for_encryption(self, address: str) -> list[tuple[str, int]]:
        devices_for_encryption: list[tuple[str, int]] = []
        devices = self.get_devices(address)

        for device in devices:
            if self._storage.is_trusted(address, device):
                devices_for_encryption.append((address, device))
        return devices_for_encryption

    def get_our_device(self) -> int:
        return self._our_device_id

    def is_our_device_published(self) -> bool:
        return self.get_our_device() in self.get_devices(self._our_address)

    def destroy(self) -> None:
        self._unregister_signals()



from __future__ import annotations

from typing import cast
from typing import Optional

import google.protobuf.message

from . import storage_pb2
from .sessionstate import SessionState
from .sessionstate import SessionStructureProto

ARCHIVED_STATES_MAX_LENGTH = 40


class SessionRecord:
    def __init__(
        self,
        session_state: Optional[SessionState] = None,
        serialized: Optional[bytes] = None,
    ) -> None:
        self._previous_states: list[SessionState] = []
        if session_state:
            self._session_state = session_state
            self._fresh = False

        elif serialized:
            record = cast(
                RecordStructureProto, storage_pb2.RecordStructure()  # pyright: ignore
            )
            record.ParseFromString(serialized)
            self._session_state = SessionState(record.currentSession)
            self._fresh = False
            for previous_structure in record.previousSessions:
                self._previous_states.append(SessionState(previous_structure))

        else:
            self._fresh = True
            self._session_state = SessionState()

    def has_session_state(self, version: int, alice_base_key: bytes) -> bool:
        if (
            self._session_state.get_session_version() == version
            and alice_base_key == self._session_state.get_alice_base_key()
        ):
            return True

        for state in self._previous_states:
            if (
                state.get_session_version() == version
                and alice_base_key == state.get_alice_base_key()
            ):
                return True

        return False

    def get_session_state(self) -> SessionState:
        return self._session_state

    def get_previous_session_states(self) -> list[SessionState]:
        return self._previous_states

    def is_fresh(self) -> bool:
        return self._fresh

    def archive_current_state(self) -> None:
        self.promote_state(SessionState())

    def promote_state(self, promoted_state: SessionState) -> None:
        self._previous_states.insert(0, self._session_state)
        self._session_state = promoted_state
        if len(self._previous_states) > ARCHIVED_STATES_MAX_LENGTH:
            self._previous_states.pop()

    def set_state(self, session_state: SessionState) -> None:
        self._session_state = session_state

    def serialize(self) -> bytes:
        previous_structures = [
            previous_state.get_structure() for previous_state in self._previous_states
        ]
        record = cast(
            RecordStructureProto, storage_pb2.RecordStructure()  # pyright: ignore
        )
        record.currentSession.MergeFrom(self._session_state.get_structure())
        record.previousSessions.extend(previous_structures)

        return record.SerializeToString()


class RecordStructureProto(google.protobuf.message.Message):
    currentSession: SessionStructureProto
    previousSessions: list[SessionStructureProto]
